import sys
from functools import wraps
from typing import (
    NewType,
    cast,
    Callable,
    Any,
    TypeVar,
    List,
    Iterator,
    Dict,
    Generic,
    NoReturn,
)

from hiku.compat import ParamSpec

Const = NewType('Const', object)


class cached_property:

    def __init__(self, func: Callable) -> None:
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj: Any, cls: Any) -> Any:
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def const(name: str) -> Const:
    t = type(name, (object,), {})
    t.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    return cast(Const, t)


T = TypeVar('T')
P = ParamSpec('P')


def listify(func: Callable[P, Iterator[T]]) -> Callable[P, List[T]]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> List[T]:
        return list(func(*args, **kwargs))
    return wrapper


K = TypeVar('K')
V = TypeVar('V')


class ImmutableDict(Dict, Generic[K, V]):
    _hash = None

    def __hash__(self) -> int:  # type: ignore
        if self._hash is None:
            self._hash = hash(frozenset(self.items()))
        return self._hash

    def _immutable(self) -> NoReturn:
        raise TypeError("{} object is immutable"
                        .format(self.__class__.__name__))

    __delitem__ = __setitem__ = _immutable  # type: ignore
    clear = pop = popitem = setdefault = update = _immutable  # type: ignore



import time
from abc import abstractmethod
from functools import partial

from prometheus_client import (
    Summary,
    Gauge,
)

from ..graph import GraphTransformer
from ..engine import pass_context, _do_pass_context
from ..sources.graph import CheckedExpr


QUERY_CACHE_HITS = Gauge('hiku_query_cache_hits', 'Query cache hits')
QUERY_CACHE_MISSES = Gauge('hiku_query_cache_misses', 'Query cache misses')


_METRIC = None


def _get_default_metric():
    global _METRIC
    if _METRIC is None:
        _METRIC = Summary(
            'graph_field_time',
            'Graph field time (seconds)',
            ['graph', 'node', 'field'],
        )
    return _METRIC


def _func_field_names(func):
    fields_pos = 1 if _do_pass_context(func) else 0

    def wrapper(*args):
        return func([f.name for f in args[fields_pos]], *args)
    return wrapper


def _subquery_field_names(func):
    def wrapper(fields, *args):
        return func([f.name for _, f in fields], fields, *args)
    return wrapper


class GraphMetricsBase(GraphTransformer):
    root_name = 'Root'

    def __init__(self, name, *, metric=None):
        self._name = name
        self._metric = metric or _get_default_metric()
        self._node = None
        self._wrappers = {}

    @abstractmethod
    def field_wrapper(self, observe, func):
        raise NotImplementedError

    @abstractmethod
    def link_wrapper(self, observe, func):
        raise NotImplementedError

    @abstractmethod
    def subquery_wrapper(self, observe, subquery):
        raise NotImplementedError

    def _observe_fields(self, node_name):
        by_field = {}

        def observe(start_time, field_names):
            duration = time.perf_counter() - start_time
            for name in field_names:
                try:
                    field_metric = by_field[name]
                except KeyError:
                    field_metric = by_field[name] = \
                        self._metric.labels(self._name, node_name, name)
                field_metric.observe(duration)
        return observe

    def _wrap_field(self, node_name, func):
        observe = self._observe_fields(node_name)
        wrapper = self.field_wrapper(observe, func)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        wrapper = _func_field_names(wrapper)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        return wrapper

    def _wrap_link(self, node_name, link_name, func):
        observe = self._observe_fields(node_name)
        wrapper = self.link_wrapper(observe, func)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        wrapper = partial(wrapper, link_name)
        if _do_pass_context(func):
            wrapper = pass_context(wrapper)
        return wrapper

    def _wrap_subquery(self, node_name, subquery):
        observe = self._observe_fields(node_name)
        wrapper = self.subquery_wrapper(observe, subquery)
        wrapper = _subquery_field_names(wrapper)
        wrapper.__subquery__ = lambda: wrapper
        return wrapper

    def visit_node(self, obj):
        self._node = obj
        try:
            return super().visit_node(obj)
        finally:
            self._node = None

    def visit_field(self, obj):
        obj = super().visit_field(obj)
        node_name = self.root_name if self._node is None else self._node.name
        if isinstance(obj.func, CheckedExpr):
            func = obj.func.__subquery__
        else:
            func = obj.func

        wrapper = self._wrappers.get(func)
        if wrapper is None:
            if isinstance(obj.func, CheckedExpr):
                wrapper = self._wrappers[func] = self._wrap_subquery(
                    node_name, func,
                )
            else:
                wrapper = self._wrappers[func] = self._wrap_field(
                    node_name, func,
                )

        if isinstance(obj.func, CheckedExpr):
            obj.func = CheckedExpr(
                wrapper,
                obj.func.expr,
                obj.func.reqs,
                obj.func.proc,
            )
        else:
            obj.func = wrapper
        return obj

    def visit_link(self, obj):
        obj = super().visit_link(obj)
        node_name = self.root_name if self._node is None else self._node.name
        obj.func = self._wrap_link(node_name, obj.name, obj.func)
        return obj


class _SubqueryMixin:

    def subquery_wrapper(self, observe, subquery):
        def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result_proc = subquery(*args)

            def proc_wrapper():
                result = result_proc()
                observe(start_time, field_names)
                return result
            return proc_wrapper
        return wrapper


class GraphMetrics(_SubqueryMixin, GraphMetricsBase):

    def field_wrapper(self, observe, func):
        def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result = func(*args)
            observe(start_time, field_names)
            return result
        return wrapper

    def link_wrapper(self, observe, func):
        def wrapper(link_name, *args):
            start_time = time.perf_counter()
            result = func(*args)
            observe(start_time, [link_name])
            return result
        return wrapper


class AsyncGraphMetrics(_SubqueryMixin, GraphMetricsBase):

    def field_wrapper(self, observe, func):
        async def wrapper(field_names, *args):
            start_time = time.perf_counter()
            result = await func(*args)
            observe(start_time, field_names)
            return result
        return wrapper

    def link_wrapper(self, observe, func):
        async def wrapper(link_name, *args):
            start_time = time.perf_counter()
            result = await func(*args)
            observe(start_time, [link_name])
            return result
        return wrapper



import typing as t

from abc import ABC, abstractmethod
from asyncio import gather
from inspect import isawaitable

from ..engine import Engine
from ..graph import (
    apply,
    Graph,
)
from ..query import (
    QueryTransformer,
    Node,
)
from ..result import Proxy
from ..validate.query import validate
from ..readers.graphql import (
    read_operation,
    OperationType,
    Operation,
)
from ..denormalize.graphql import DenormalizeGraphQL
from ..introspection.graphql import AsyncGraphQLIntrospection, QUERY_ROOT_NAME
from ..introspection.graphql import GraphQLIntrospection, MUTATION_ROOT_NAME


_type_names: t.Dict[OperationType, str] = {
    OperationType.QUERY: QUERY_ROOT_NAME,
    OperationType.MUTATION: MUTATION_ROOT_NAME,
}


class GraphQLError(Exception):

    def __init__(self, *, errors: t.List[str]):
        super().__init__('{} errors'.format(len(errors)))
        self.errors = errors


class _StripQuery(QueryTransformer):
    """Removes __typename fields from query"""

    def visit_node(self, obj: Node) -> Node:
        return obj.copy(fields=[self.visit(f) for f in obj.fields
                                if f.name != '__typename'])


def _switch_graph(
    data: t.Dict,
    query_graph: Graph,
    mutation_graph: t.Optional[Graph] = None
) -> t.Tuple[Graph, Operation]:
    try:
        op = read_operation(data['query'],
                            variables=data.get('variables'),
                            operation_name=data.get('operationName'))
    except TypeError as e:
        raise GraphQLError(errors=[
            'Failed to read query: {}'.format(e),
        ])
    if op.type is OperationType.QUERY:
        graph = query_graph
    elif op.type is OperationType.MUTATION and mutation_graph is not None:
        graph = mutation_graph
    else:
        raise GraphQLError(errors=[
            'Unsupported operation type: {!r}'.format(op.type),
        ])
    return graph, op


def _process_query(graph: Graph, query: Node) -> Node:
    stripped_query = _StripQuery().visit(query)
    errors = validate(graph, stripped_query)
    if errors:
        raise GraphQLError(errors=errors)
    else:
        return stripped_query


class BaseGraphQLEndpoint(ABC):
    query_graph: Graph
    mutation_graph: t.Optional[Graph]

    @property
    @abstractmethod
    def introspection_cls(self) -> t.Type[GraphQLIntrospection]:
        pass

    def __init__(
        self,
        engine: Engine,
        query_graph: Graph,
        mutation_graph: t.Optional[Graph] = None
    ):
        self.engine = engine

        introspection = self.introspection_cls(query_graph, mutation_graph)
        self.query_graph = apply(query_graph, [introspection])
        if mutation_graph is not None:
            self.mutation_graph = apply(mutation_graph, [introspection])
        else:
            self.mutation_graph = None


class BaseSyncGraphQLEndpoint(BaseGraphQLEndpoint):
    @abstractmethod
    def execute(self, graph: Graph, op: Operation, ctx: t.Dict) -> t.Dict:
        pass

    @abstractmethod
    def dispatch(self, data: t.Dict) -> t.Dict:
        pass


class BaseAsyncGraphQLEndpoint(BaseGraphQLEndpoint):
    @abstractmethod
    async def execute(self, graph: Graph, op: Operation, ctx: t.Dict) -> t.Dict:
        pass

    @abstractmethod
    async def dispatch(self, data: t.Dict) -> t.Dict:
        pass


class GraphQLEndpoint(BaseSyncGraphQLEndpoint):
    introspection_cls = GraphQLIntrospection

    def execute(
        self, graph: Graph, op: Operation, ctx: t.Optional[t.Dict]
    ) -> t.Dict:
        stripped_query = _process_query(graph, op.query)
        result = self.engine.execute(graph, stripped_query, ctx, op)
        assert isinstance(result, Proxy)
        type_name = _type_names[op.type]
        return DenormalizeGraphQL(graph, result, type_name).process(op.query)

    def dispatch(self, data: t.Dict) -> t.Dict:
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )
            result = self.execute(graph, op, {})
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class BatchGraphQLEndpoint(GraphQLEndpoint):

    @t.overload
    def dispatch(self, data: t.List[t.Dict]) -> t.List[t.Dict]:
        ...

    @t.overload
    def dispatch(self, data: t.Dict) -> t.Dict:
        ...

    def dispatch(
        self, data: t.Union[t.Dict, t.List[t.Dict]]
    ) -> t.Union[t.Dict, t.List[t.Dict]]:
        if isinstance(data, list):
            return [
                super(BatchGraphQLEndpoint, self).dispatch(item)
                for item in data
            ]
        else:
            return super(BatchGraphQLEndpoint, self).dispatch(data)


class AsyncGraphQLEndpoint(BaseAsyncGraphQLEndpoint):
    introspection_cls = AsyncGraphQLIntrospection

    async def execute(
        self, graph: Graph, op: Operation, ctx: t.Optional[t.Dict]
    ) -> t.Dict:
        stripped_query = _process_query(graph, op.query)
        coro = self.engine.execute(graph, stripped_query, ctx, op)
        assert isawaitable(coro)
        result = await coro
        type_name = _type_names[op.type]
        return DenormalizeGraphQL(graph, result, type_name).process(op.query)

    async def dispatch(self, data: t.Dict) -> t.Dict:
        try:
            graph, op = _switch_graph(
                data, self.query_graph, self.mutation_graph,
            )
            result = await self.execute(graph, op, {})
            return {'data': result}
        except GraphQLError as e:
            return {'errors': [{'message': e} for e in e.errors]}


class AsyncBatchGraphQLEndpoint(AsyncGraphQLEndpoint):

    @t.overload
    async def dispatch(self, data: t.List[t.Dict]) -> t.List[t.Dict]:
        ...

    @t.overload
    async def dispatch(self, data: t.Dict) -> t.Dict:
        ...

    async def dispatch(
        self, data: t.Union[t.Dict, t.List[t.Dict]]
    ) -> t.Union[t.Dict, t.List[t.Dict]]:
        if isinstance(data, list):
            return list(await gather(*(
                super(AsyncBatchGraphQLEndpoint, self).dispatch(item)
                for item in data
            )))
        else:
            return await super(AsyncBatchGraphQLEndpoint, self).dispatch(data)



from typing import (
    Optional,
    Union,
    TYPE_CHECKING,
    Any,
)

if TYPE_CHECKING:
    from hiku.graph import Field, Link


class DirectiveBase:
    pass


class Deprecated(DirectiveBase):
    """
    https://spec.graphql.org/June2018/#sec--deprecated
    """
    def __init__(self, reason: Optional[str] = None):
        self.reason = reason

    def accept(self, visitor: Any) -> Any:
        return visitor.visit_deprecated_directive(self)


def get_deprecated(field: Union['Field', 'Link']) -> Optional[Deprecated]:
    """Get deprecated directive"""
    return next(
        (d for d in field.directives if isinstance(d, Deprecated)),
        None
    )


class QueryDirective:
    def __init__(self, name: str) -> None:
        self.name = name


class Cached(QueryDirective):
    def __init__(self, ttl: int):
        super().__init__('cached')
        self.ttl = ttl
